{-# LANGUAGE OverloadedStrings #-}

module Lisa.Squeak where

import Control.Applicative ((<|>))
import Data.Aeson (FromJSON(parseJSON), ToJSON(toJSON), Value, (.:), (.:?), (.=), object, withObject)
import Data.Text (Text)
import Data.Time.Calendar (Day)
import Network.HTTP.Req (POST(..), ReqBodyJson(..), Req, http, https, port, jsonResponse, responseBody, req)

data GQLQuery = GQLQuery
    { gqlQueryQuery     :: Text
    }
    deriving (Show)

instance ToJSON GQLQuery where
    toJSON (GQLQuery query) = object
        [ "query" .= query
        ]

data GQLQueryWithVars a = GQLQueryWithVars
    { gqlQueryWithVarsQuery     :: Text
    , gqlQueryWithVarsVariables :: a
    }
    deriving (Show)

instance ToJSON a => ToJSON (GQLQueryWithVars a) where
    toJSON (GQLQueryWithVars query variables) = object
        [ "query" .= query
        , "variables" .= variables
        ]

data GQLReply a = GQLReply
    { gqlReplyData :: Maybe a
    , gqlReplyErrors :: Maybe [GQLError]
    }
    deriving (Show)

instance FromJSON a => FromJSON (GQLReply a) where
    parseJSON = withObject "GQLReply" $ \v -> GQLReply
        <$> v .: "data"
        <*> v .:? "errors"

replyToEither :: GQLReply a -> Either [GQLError] a
replyToEither (GQLReply (Just data') _)  = Right data'
replyToEither (GQLReply _ (Just errors)) = Left errors
replyToEither _                          = error "Invalid GQLReply: Neither 'data' nor 'errors'"

data GQLError = GQLError
    { gqlErrorMessage :: Text
    }
    deriving (Show)

instance FromJSON GQLError where
    parseJSON = withObject "GQLError" $ \v -> GQLError
        <$> v .: "message"

data Documents = Documents { unDocuments :: [Document] }
    deriving (Show)

instance FromJSON Documents where
    parseJSON = withObject "Documents" $ \v -> Documents
        <$> ((v .: "documents") >>= (.: "results"))

data Document = Document
    { documentId            :: Text
    , documentDate          :: Day
    , documentSemester      :: Text
    , documentPublicComment :: Maybe Text
    , documentFaculty       :: Faculty
    , documentLectures      :: [Lecture]
    , documentDownloadable  :: Bool
    }
    deriving (Show)

instance FromJSON Document where
    parseJSON = withObject "Document" $ \v -> Document
        <$> v .: "id"
        <*> v .: "date"
        <*> v .: "semester"
        <*> v .: "publicComment"
        <*> v .: "faculty"
        <*> v .: "lectures"
        <*> v .: "downloadable"

data Faculty = Faculty
    { facultyId          :: Text
    , facultyDisplayName :: Text
    }
    deriving (Show)

instance FromJSON Faculty where
    parseJSON = withObject "Faculty" $ \v -> Faculty
        <$> v .: "id"
        <*> v .: "displayName"

data Lectures = Lectures { unLectures :: [Lecture] }
    deriving (Show)

instance FromJSON Lectures where
    parseJSON = withObject "Lectures" $ \v -> Lectures
        <$> v .: "lectures"

data Lecture = Lecture
    { lectureId          :: Text
    , lectureDisplayName :: Text
    , lectureAliases     :: [Text]
    }
    deriving (Show)

instance FromJSON Lecture where
    parseJSON = withObject "Lecture" $ \v -> Lecture
        <$> v .: "id"
        <*> v .: "displayName"
        <*> v .: "aliases"

serverUrl = https "api.squeak-test.fsmi.uni-karlsruhe.de"

getAllLectures :: Req (Either [GQLError] [Lecture])
getAllLectures = fmap unLectures <$> replyToEither <$> responseBody <$> req POST serverUrl (ReqBodyJson $ GQLQuery q) jsonResponse mempty
    where q = "{ lectures { id displayName aliases } }"

getAllDocuments :: Req (Either [GQLError] [Document])
getAllDocuments = fmap unDocuments <$> replyToEither <$> responseBody <$> req POST serverUrl (ReqBodyJson $ GQLQuery q) jsonResponse mempty
    where q = "{ documents(filters: []) { results { id date semester publicComment downloadable faculty { id displayName } lectures { id displayName aliases } } } }"

-- Mutations

data LoginResult = LoginResult { unLoginResult :: Credentials }
    deriving (Show)

instance FromJSON LoginResult where
    parseJSON = withObject "LoginResult" $ \v -> LoginResult
        <$> v .: "login"

data Credentials = Credentials
    { credentialsToken :: Text
    , credentialsUser  :: User
    } deriving (Show)

instance FromJSON Credentials where
    parseJSON = withObject "Credentials" $ \v -> Credentials
        <$> v .: "token"
        <*> v .: "user"

data User = User
    { userUsername    :: Text
    , userDisplayName :: Text
    } deriving (Show)

instance FromJSON User where
    parseJSON = withObject "User" $ \v -> User
        <$> v .: "username"
        <*> v .: "displayName"

login :: Text -> Text -> Req (Either [GQLError] Credentials)
login username password = fmap unLoginResult <$> replyToEither <$> responseBody <$> req POST serverUrl (ReqBodyJson $ GQLQueryWithVars q vars) jsonResponse mempty
    where q = "mutation LoginUser($username: String!, $password: String!) { login(username: $username, password: $password) { ... on Credentials { token user { username displayName } } } }"
          vars = object
            [ "username" .= username
            , "password" .= password
            ]

data SqueakError
    = SqueakError
        Text -- ^ Error code
        Text -- ^ Human-readable message
    | GQLErrors [GQLError]
    deriving (Show)

instance FromJSON SqueakError where
    parseJSON = withObject "SqueakError" $ \v -> SqueakError
        <$> v .: "errorCode"
        <*> v .: "msg"

newtype SqueakErrorOr a = SqueakErrorOr (Either SqueakError a)
    deriving (Show)

instance FromJSON a => FromJSON (SqueakErrorOr a) where
    -- | Parse 'SqueakError' if @errorCode@ and @msg@ are present; otherwise, parse an 'a'.
    parseJSON v = SqueakErrorOr
        <$> ((Left <$> parseJSON v)
        <|> (Right <$> parseJSON v))

data CreateOrderResult = CreateOrderResult { unCreateOrderResult :: SqueakErrorOr Order }
    deriving (Show)

instance (FromJSON CreateOrderResult) where
    parseJSON = withObject "CreateOrderResult" $ \v -> CreateOrderResult
        <$> v .: "createOrder"

data Order = Order
    { orderId       :: Text
    , orderTag      :: Text
    , orderNumPages :: Int
    }
    deriving (Show)

instance FromJSON Order where
    parseJSON = withObject "Order" $ \v -> Order
        <$> v .: "id"
        <*> v .: "tag"
        <*> v .: "numPages"

joinErrors :: Either [GQLError] (SqueakErrorOr a) -> Either SqueakError a
joinErrors (Left errors) = Left $ GQLErrors errors
joinErrors (Right (SqueakErrorOr inner)) = inner

createOrder :: Text -> [Text] -> Req (Either SqueakError Order)
createOrder tag documents = joinErrors <$> fmap unCreateOrderResult <$> replyToEither <$> responseBody <$> req POST serverUrl (ReqBodyJson $ GQLQueryWithVars q vars) jsonResponse mempty
    where q = "mutation CreateOrder($tag: String!, $documents: [DocumentId!]!) \
              \ { createOrder(tag: $tag, documents: $documents) \
              \ { ... on Error { errorCode msg } \
              \   ... on Order { id tag numPages } \
              \ } }"
          vars = object
            [ "tag" .= tag
            , "documents" .= documents
            ]
